/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           MSX.c                         **/
/**                                                         **/
/** This file contains implementation for the MSX-specific  **/
/** hardware: slots, memory mapper, PPIs, VDP, PSG, clock,  **/
/** etc. Initialization code and definitions needed for the **/
/** machine-dependent drivers are also here.                **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "MSX.h"
#include "Boot.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#define PRINTOK     if(Verbose) printf(MsgOK)
#define PRINTFAILED if(Verbose) printf(MsgFAILED)

char *MsgOK      = "OK\n";
char *MsgFAILED  = "FAILED\n";

byte Verbose     = 1;               /* Debug msgs ON/OFF      */
byte SaveCMOS    = 0;               /* Save CMOS.ROM on exit  */
byte MSXVersion  = 0;               /* 0=MSX1,1=MSX2,2=MSX2+  */
byte ROMType     = 0;               /* MegaROM type           */ 
byte UPeriod     = 2;               /* Interrupts/scr. update */
byte RAMPages    = 4;               /* Number of RAM pages    */
byte VRAMPages   = 4;               /* Number of VRAM pages   */

char *CartA      = "CARTA.ROM";     /* Cartridge A ROM file   */
char *CartB      = "CARTB.ROM";     /* Cartridge B ROM file   */

char *DiskA      = "DRIVEA.DSK";    /* Drive A disk image     */
char *DiskB      = "DRIVEB.DSK";    /* Drive B disk image     */
int  Drives[2]   = { -1,-1 };       /* Disk image files       */

char *FontName   = "DEFAULT.FNT";   /* Font file for text     */
byte *FontBuf    = NULL;            /* Font for text modes    */

char *PrnName    = NULL;            /* Printer redirect. file */
FILE *PrnStream  = NULL;

byte *VRAM=NULL,*RAM=NULL,*VPAGE;   /* Main and Video RAMs    */

byte *MemMap[4][4][4];   /* Memory maps [PPage][SPage][Addr]  */

byte *RAMMap[16];                   /* Blocks for RAM Mapper  */
byte RAMMapper[4];                  /* RAM Mapper state       */
byte *ROMMap[2][256];               /* Blocks for ROM Mappers */
byte ROMMapper[2][4];               /* ROM Mappers state      */ 

byte EnWrite[4] = {0,0,0,0};        /* 1 if write enabled     */
byte CopyOut[4] = {0,0,0,0};        /* 1 if page copied out   */

byte PSL[4],SSL[4];                 /* Lists of current slots */
byte PSLReg,SSLReg;      /* Storage for A8h port and (FFFFh)  */

byte KeyMap[16] =                   /* Keyboard map           */
{
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

byte IOReg;                         /* Storage for AAh port   */
byte PPIReg;                        /* PPI control register   */

byte OPLLReg=0;                     /* OPLL register number   */
byte PSGReg=0;                      /* PSG register number    */ 
byte OPLL[16];                      /* OPLL registers         */
byte PSG[16];                       /* PSG registers          */
byte SCC[16];                       /* SCC registers          */

byte RTCReg=0,RTCMode=0;            /* RTC register numbers   */
byte RTC[4][13];                    /* RTC registers          */

byte *ChrGen,*ChrTab,*ColTab;       /* VDP tables [screens]   */
byte *SprGen,*SprTab,*SprCol;       /* VDP tables [sprites]   */
byte Palette[32][3];                /* Palette registers      */
pair WVAddr,RVAddr;      /* Storage for VRAM addresses in VDP */
byte VKey=1,PKey=1;                 /* Status keys for VDP    */
byte FGColor=0,BGColor=0;           /* Colors                 */
byte XFGColor,XBGColor;             /* Second set of colors   */
byte ScrMode=0;                     /* Current screen mode    */
byte VDP[64],VDPStatus[16];         /* VDP registers          */
byte ScanLine=0;                    /* Current scanline       */
byte EndOfFrame=1;                  /* 1 when end of frame    */

/******** Places in DiskROM to be patched with ED FE C9: ******/
word DiskPatches[] = { 0x4010,0x4013,0x401C,0 };

/******** Screen Mode Handlers [number of screens + 1]: *******/
extern struct
{                                   /* Screen mode handlers:  */
  int (*Init)();                    /* Initialize screen      */
  void (*Colors)();                 /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2); /* Refresh screen         */
  byte R2,R3,R4,R5;                 /* Masks for table addr.  */
} SCR[MAXSCREEN+2];                 /* Number of screens + 1  */


/********** Data for the CPU->VRAM function of V9938: *********/
struct
{
  byte *P,Mode,TR,Mask,Op,Diff,Shift[4];
  int X,NX,NY,XC,NXC,XStep,YStep;
} MMC;


/* Load a .ROM file into given slot */
byte *LoadROM(char *STR,FILE *F,int P,int S,int A);
int LoadCart(char *Name,int Slot);/* Load cartridge from a .ROM file */
void VDPOut(byte Reg,byte Value); /* Write value into a VDP register */
void Printer(byte V);             /* Send a character to a printer   */
void CheckSprites(void);      /* Check for collisions and 5th sprite */
byte RTCIn(byte R);               /* Read RTC registers              */
void DoVDP(byte V);               /* Perform V9938 operation         */
/* Function used by DoVDP() for operations on pixels */ 
byte Oper(byte A,byte B,byte N);


/****************************************************************/
/*** Write number into given IO port.                         ***/
/****************************************************************/
void DoOut(register byte Port,register byte Value)
{
  static byte Pal,Addr;     /* Palette and address buffers */
  register byte I,J,K,*P;  

  switch(Port)
  {

case 0x7C: OPLLReg=Value&0x0F;return;
case 0x7D: OPLL[OPLLReg]=Value;
           OPLLOut(OPLLReg,Value);
           return;
case 0x91: Printer(Value);return;
case 0x98: if(VKey)
           {
             if(VPAGE) *(VPAGE+WVAddr.W)=Value;
             WVAddr.W=(WVAddr.W+1)&0x3FFF;
             if(!WVAddr.W&&(ScrMode>3)) 
             {
               VDP[14]=(VDP[14]+1)&0x07;
               VPAGE=(VDP[14]<VRAMPages)?
                 VRAM+((long)VDP[14]<<14):NULL;
             }
           }
           return;
case 0x99: if(VKey) { Addr=Value;VKey--; }
           else
           {
             VKey++;
             switch(Value&0xC0)
             {
               case 0x80: VDPOut(Value&0x3F,Addr);break;
               case 0x40: WVAddr.B.l=Addr;WVAddr.B.h=Value&0x3F;break;
               case 0x00: RVAddr.B.l=Addr;RVAddr.B.h=Value;
             }
           }
           return;
case 0x9A: if(PKey) { Pal=Value;PKey--; }
           else
           {
             PKey++;J=VDP[16];
             VDP[16]=(VDP[16]+1)&0x0F;
             Palette[J][0]=(Pal&0x70)<<1;
             Palette[J][1]=(Value&0x07)<<5;
             Palette[J][2]=(Pal&0x07)<<5;
             SCR[ScrMode].Colors();
           }
           return;
case 0x9B: J=VDP[17]&0x3F;
           if(J!=17) VDPOut(J,Value);
           if(!(VDP[17]&0x80)) VDP[17]=(J+1)&0x3F;
           return;
case 0xA0: PSGReg=Value&0x0F;return;
case 0xA1: PSG[PSGReg]=Value;
           PSGOut(PSGReg,Value);
           return;
case 0xAA: IOReg=Value;return;
case 0xAB: if(Value&0x80) PPIReg=Value;
           else
             if(Value&0x01) IOReg|=1<<(7&(Value>>1));
             else           IOReg&=~(1<<(7&(Value>>1)));
           return;
case 0xB4: RTCReg=Value&0x0F;return;
case 0xB5: if(RTCReg<13)
           {
             J=RTCMode&0x03;RTC[J][RTCReg]=Value;
             if(J>1) SaveCMOS=1;
           }
           else if(RTCReg==13) RTCMode=Value;
           return;
case 0xA8: if(Value!=PSLReg)
           {
             for(J=0,PSLReg=Value;J<4;J++,Value>>=2)
             {
               I=Value&3;
               if(PSL[J]!=I)
               {
                 P=RAM+((word)J<<14);K=SSL[J];
                 if(CopyOut[J])
                   memcpy(MemMap[PSL[J]][K][J],P,PAGESIZE);
                 PSL[J]=I;
                 if(MemMap[I][K][J])
                   memcpy(P,MemMap[I][K][J],PAGESIZE); 
                 else
                   memset(P,NORAM,PAGESIZE);
                 CopyOut[J]=EnWrite[J]=(I==3)&&(K==2)&&MemMap[3][2][J];
               }  
             }
             if(PSL[3]==3) *(RAM+0xFFFF)=~SSLReg;
           }
           return;
case 0xFC:
case 0xFD:
case 0xFE:
case 0xFF: I=Port-0xFC;Value&=0x0F;
           if((PSL[I]==3)&&(SSL[I]==2))
           {
             P=RAM+((word)I<<14);
             if(EnWrite[I]) memcpy(MemMap[3][2][I],P,PAGESIZE);
             EnWrite[I]=RAMMap[Value]? 1:0;
             if(EnWrite[I]) memcpy(P,RAMMap[Value],PAGESIZE);
             else           memset(P,NORAM,PAGESIZE);
           }
           MemMap[3][2][I]=RAMMap[Value];RAMMapper[I]=Value;
           return; 

  }
}



/****************************************************************/
/*** Read number from given IO port.                          ***/
/****************************************************************/
byte DoIn(register byte Port)
{
  switch(Port)
  {

case 0x90: return(0xFD);                     /* Printer READY signal */
case 0x98: Port=VPAGE? *(VPAGE+RVAddr.W):255;/* VRAM read port       */
           RVAddr.W=(RVAddr.W+1)&0x3FFF;
           if(!RVAddr.W&&(ScrMode>3))
           {
             VDP[14]=(VDP[14]+1)&0x07;
             VPAGE=(VDP[14]<VRAMPages)?
               VRAM+((long)VDP[14]<<14):NULL;
           }
           return(Port);
case 0x99: Port=VDPStatus[VDP[15]];          /* VDP status registers */
           switch(VDP[15])
           {
             case 0: VDPStatus[0]&=0x5F;break;
             case 1: VDPStatus[1]&=0xFE;break;
             case 2: VDPStatus[2]^=0x60;break;
           }
           return(Port);
case 0xA2: if(PSGReg>13) return(NORAM);      /* PSG registers        */
           else return(PSG[PSGReg]);
case 0xA8: return(PSLReg);                   /* Primary slot state   */
case 0xA9: return(KeyMap[IOReg&0x0F]);       /* Keyboard port        */
case 0xAA: return(IOReg);                    /* General IO register  */
case 0xAB: return(PPIReg);                   /* PPI control register */
case 0xB5: return(RTCIn(RTCReg));            /* RTC registers        */
case 0xFC:                                   /* Mapper page at 0000h */
case 0xFD:                                   /* Mapper page at 4000h */
case 0xFE:                                   /* Mapper page at 8000h */
case 0xFF: return(RAMMapper[Port-0xFC]|0xF0);/* Mapper page at C000h */
default:   return(NORAM);                    /* If no port, ret FFh  */ 

  }
}



/****************************************************************/
/*** Switch ROM Mapper pages. This function is supposed to be ***/
/*** to be called when ROM page registers are written to.     ***/
/****************************************************************/
void MapROM(register word Addr,register byte Value)
{
  byte I,J,K;

  J=Addr>>14;

  if((PSL[J]==1)&&ROMMap[0][0]) I=0;
  else 
    if((PSL[J]==2)&&ROMMap[1][0]) I=1;
    else return;

  if(MemMap[I+1][0][J])
  {
    switch(ROMType)
    {
      case 0: /*** KONAMI 8kB cartridges ***/
        if((ROMMapper[I][3]==0x3F)&&((Addr&0xFF00)==0x9800))
        { J=Addr&0x000F;SCC[J]=Value;SCCOut(J,Value);return; }
        Addr=(Addr-0x4000)&0xE000;J=Addr>>13;
        K=0;break;
      case 1: /*** KONAMI 16kB cartridges ***/
        Addr=(Addr&0x8000)>>1;
        J=Addr>>13;Value<<=1;
        K=1;break;
      case 2: /*** ASCII 8kB cartridges ***/
        if((Addr>=0x8000)||(Addr<0x6000)) return;
        Addr=(Addr&0x1800)<<2;J=Addr>>13;
        K=0;break;
      case 3: /*** ASCII 16kB cartridges ***/
        if((Addr>=0x8000)||(Addr<0x6000)) return;
        Addr=(Addr&0x1000)<<2;
        J=Addr>>13;Value<<=1;
        K=1;break;
      default:
        printf("Unknown MegaROM type %d\n",ROMType);
        return; 
    }

    if(Value==ROMMapper[I][J]) return;
    ROMMapper[I][J]=Value;CopyOut[(J>>1)+1]=1;
    Addr+=0x4000;

    if(K)
    {
      if(Verbose&0x08)
        printf("ROM-MAPPER: 16kB block %d at %Xh\n",Value,Addr);
      ROMMapper[I][J+1]=Value+1;
      if(ROMMap[I][Value])
      {
        memcpy(RAM+Addr,ROMMap[I][Value],PAGESIZE/2);
        memcpy(RAM+Addr+0x2000,ROMMap[I][Value+1],PAGESIZE/2);
      }
      else memset(RAM+Addr,NORAM,PAGESIZE);
    }
    else
    {
      if(Verbose&0x08)
        printf("ROM-MAPPER: 8kB block %d at %Xh\n",Value,Addr);
      if(ROMMap[I][Value])
        memcpy(RAM+Addr,ROMMap[I][Value],PAGESIZE/2);
      else
        memset(RAM+Addr,NORAM,PAGESIZE/2);
    }
  }
}



/****************************************************************/
/*** Switch secondary memory slots. This function is supposed ***/
/*** to be called when value in (FFFFh) changes.              ***/
/****************************************************************/
void SSlot(register byte Value)
{
  register byte I,J,K,*P;
  
  if(PSL[3]!=3)
  { if(EnWrite[3]) *(RAM+0xFFFF)=Value;return; }

  if(SSLReg!=Value)
  {
    SSLReg=Value;

    for(J=0;J<4;J++,Value>>=2)
    {
      I=Value&3;K=SSL[J];SSL[J]=I;
      if((K!=I)&&(PSL[J]==3))
      {
        P=RAM+((word)J<<14);
        if(CopyOut[J])      memcpy(MemMap[3][K][J],P,PAGESIZE);
        if(MemMap[3][I][J]) memcpy(P,MemMap[3][I][J],PAGESIZE);
        else                memset(P,NORAM,PAGESIZE);
        CopyOut[J]=EnWrite[J]=(I==2)&&MemMap[3][I][J];
      }
    }
    *(RAM+0xFFFF)=~SSLReg;
  }
}
  
  

/****************************************************************/
/*** Allocate memory, load ROM images, initialize mapper, VDP ***/
/*** CPU and start the emulation. This function returns 0 in  ***/
/*** the case of failure.                                     ***/
/****************************************************************/
int StartMSX(void)
{
  word A;
  FILE *F;
  reg R;
  int *T,I,J,K;
  byte *P;

  /*** MSX versions:                ***/
  static char *Versions[] = { "MSX","MSX2","MSX2+" };
  /*** CMOS ROM default values:     ***/ 
  static byte RTCInit[4][13]  = 
  {
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 },
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 },
    { 0,0,0,0,40,80,15,4,4,0,0,0,0 },
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 }
  };
  /*** PSG initial register states: ***/
  static byte PSGInit[16]  = { 0,0,0,0,0,0,0,0xFD,0,0,0,0,0,0,0xFF,0 };
  /*** VDP status register states:  ***/
  static byte VDPSInit[16] = { 0x80,0,0x4C,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  /*** VDP control register states: ***/
  static byte VDPInit[64]  =
  {
    0x00,0x10,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  };
  /*** Initial palette: ***/
  static byte PalInit[16][3] =
  {
    {0x00,0x00,0x00},{0x00,0x00,0x00},{0x20,0xC0,0x20},{0x60,0xE0,0x60},
    {0x20,0x20,0xE0},{0x40,0x60,0xE0},{0xA0,0x20,0x20},{0x40,0xC0,0xE0},
    {0xE0,0x20,0x20},{0xE0,0x60,0x60},{0xC0,0xC0,0x20},{0xC0,0xC0,0x80},
    {0x20,0x80,0x20},{0xC0,0x40,0xA0},{0xA0,0xA0,0xA0},{0xE0,0xE0,0xE0},
  };

  /*** STARTUP CODE starts here: ***/

  T=(int *)"\01\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#ifdef LSB_FIRST
  if(*T!=1)
  {
    printf("********** This machine is high-endian. **********\n");
    printf("Take #define LSB_FIRST out and compile fMSX again.\n");
    return(0);
  }
#else
  if(*T==1)
  {
    printf("********* This machine is low-endian. **********\n");
    printf("Insert #define LSB_FIRST and compile fMSX again.\n");
    return(0);
  }
#endif

  if(MSXVersion>2) MSXVersion=2;
  if(ROMType>3) ROMType=0;
  if((RAMPages<(MSXVersion? 8:4))||(RAMPages>16))
    RAMPages=MSXVersion? 8:4;
  if((VRAMPages<(MSXVersion? 4:2))||(VRAMPages>8))
    VRAMPages=MSXVersion? 8:2;

  for(I=0;I<16;I++)  RAMMap[I]=NULL;
  for(I=0;I<256;I++) ROMMap[0][I]=ROMMap[1][I]=NULL;

  for(I=0;I<4;I++)
    for(J=0;J<4;J++)
      for(K=0;K<4;K++)
        MemMap[I][J][K]=NULL;

  if(Verbose) printf("Allocating 64kB for address space...");
  if(!(RAM=malloc(4*PAGESIZE))) return(0);
  if(Verbose) printf("OK\nAllocating %dkB for VRAM...",VRAMPages*16);
  if(!(VRAM=malloc(VRAMPages*PAGESIZE))) return(0);
  if(Verbose) printf("OK\nAllocating %dx16kB RAM pages...",RAMPages);
  for(J=0;J<RAMPages;J++) 
    if(!(RAMMap[J]=malloc(PAGESIZE))) return(0);

  if(Verbose) printf("OK\nLoading %s ROMs:\n",Versions[MSXVersion]);
    
  switch(MSXVersion)
  {
    case 0:
      if(Verbose) printf("  Opening MSX.ROM...");
      F=fopen("MSX.ROM","rb");
      break;
    case 1:
      if(Verbose) printf("  Opening MSX2EXT.ROM...");
      if(F=fopen("MSX2EXT.ROM","rb"))
      {
        if(LoadROM("OK\n    MSX2 ROM Extension",F,3,1,0)) J=1;
        fclose(F);if(!J) return(0);   
      }
      else { PRINTFAILED;return(0); }
      if(Verbose) printf("  Opening MSX2.ROM...");
      F=fopen("MSX2.ROM","rb");
      break;
    case 2:
      if(Verbose) printf("  Opening MSX2PEXT.ROM...");
      if(F=fopen("MSX2PEXT.ROM","rb"))
      {
        if(LoadROM("OK\n    MSX2+ ROM Extension",F,3,1,0)) J=1;
        fclose(F);if(!J) return(0);
      }
      else { PRINTFAILED;return(0); }
      if(Verbose) printf("  Opening MSX2P.ROM...");
      F=fopen("MSX2P.ROM","rb");
      break;
  }
  if(F)  
  {
    if(Verbose) printf("OK\n");
    J=0;
    if(LoadROM("    BIOS  ROM",F,0,0,0))
      if(LoadROM("    BASIC ROM",F,0,0,1)) J=1;
    fclose(F);if(!J) return(0);J=0;
  }
  else { PRINTFAILED;return(0); }

  if(F=fopen("CMOS.ROM","rb"))
  {
    if(Verbose) printf("Found CMOS.ROM: loading...");
    if(fread(RTC,1,sizeof(RTC),F)!=sizeof(RTC))
    { PRINTFAILED;fclose(F);F=0; }
  }
  if(F) { fclose(F);PRINTOK; }
  else memcpy(RTC,RTCInit,sizeof(RTC));

  if(F=fopen("DISK.ROM","rb"))
  {
    if(LoadROM("Found DISK.ROM",F,3,1,1))
    {
      if(Verbose) printf("Patching DiskROM: ");
      for(J=0;DiskPatches[J];J++)
      {
        if(Verbose) printf("%X..",DiskPatches[J]);
        P=MemMap[3][1][1]+DiskPatches[J]-0x4000;
        P[0]=0xED;P[1]=0xFE;P[2]=0xC9;
      }
      if(Verbose) printf(".OK\n");
    }
    fclose(F);
  }

  if(F=fopen("RS232.ROM","rb"))
  { LoadROM("Found RS232.ROM",F,3,3,1);fclose(F); }

  LoadCart(CartA,1);LoadCart(CartB,2);

  if(F=fopen(FontName,"rb"))
  {
    if(Verbose) printf("Using %s font: allocating memory...",FontName);
    if(FontBuf=malloc(2048))
    {
      if(Verbose) printf("loading...");
      if(fread(FontBuf,1,2048,F)!=2048) { free(FontBuf);FontBuf=NULL; }
    }
    if(Verbose) printf(FontBuf? MsgOK:MsgFAILED);
    fclose(F);
  }

  if(!PrnName) PrnStream=stdout;
  else
  {
    if(Verbose) printf("Redirecting printer output to %s...",PrnName);
    if(!(PrnStream=fopen(PrnName,"wb"))) PrnStream=stdout;
    printf((PrnStream==stdout)? MsgFAILED:MsgOK);
  }

  if((Drives[0]=open(DiskA,O_RDWR))>=0)
    if(Verbose) printf("Using %s as drive A\n",DiskA);  
  if((Drives[1]=open(DiskB,O_RDWR))>=0)
    if(Verbose) printf("Using %s as drive B\n",DiskB);

  if(Verbose) printf("Initializing memory mappers...");
  for(PSLReg=0x00,J=0;J<4;J++)
  { 
    PSL[J]=0;MemMap[3][2][J]=RAMMap[J];
    ROMMapper[0][J]=ROMMapper[1][J]=RAMMapper[J]=J;
  }
  for(SSLReg=0x00,J=0;J<4;J++) SSL[J]=0;  
  memcpy(RAM,MemMap[0][0][0],PAGESIZE);
  memcpy(RAM+0x4000,MemMap[0][0][1],PAGESIZE);
  memset(RAM+0x8000,NORAM,2*PAGESIZE);

  if(Verbose) printf("OK\nInitializing VDP, PSG, and CPU...");
  memcpy(VDP,VDPInit,sizeof(VDP));
  memcpy(VDPStatus,VDPSInit,sizeof(VDPStatus));
  memcpy(Palette,PalInit,sizeof(Palette));
  memcpy(PSG,PSGInit,sizeof(PSG));

  VKey=PKey=1;
  ChrTab=ColTab=ChrGen=SprTab=SprGen=VPAGE=VRAM;
  FGColor=BGColor=ScrMode=0;
  R.PC.W=0;R.SP.W=0xF000;R.IFF=0;

  if(Verbose) printf("OK\nRUNNING ROM CODE...\n");
  A=Z80(RAM,R);

  if(Verbose) printf("EXITED at PC = %Xh.\n",A);
  return(1);
}



/****************************************************************/
/*** Free memory allocated with StartMSX().                   ***/
/****************************************************************/
void TrashMSX(void)
{
  byte I,J,K;
  FILE *F;
  
  if(RAM)  free(RAM);
  if(VRAM) free(VRAM);
  if(FontBuf) free(FontBuf);

  for(I=0;I<4;I++)
    for(J=0;J<((I==3)? 4:1);J++)
      for(K=0;K<4;K++)
        if(MemMap[I][J][K]) free(MemMap[I][J][K]);

  if(SaveCMOS)
  {
    if(Verbose) printf("Writing CMOS.ROM...");
    if(F=fopen("CMOS.ROM","wb"))
      if(fwrite(RTC,1,sizeof(RTC),F)!=sizeof(RTC))
      { fclose(F);F=0; }
    if(F) { fclose(F);PRINTOK; }
    else PRINTFAILED;
  }

  if(PrnStream&&(PrnStream!=stdout)) fclose(PrnStream);
  if(Drives[0]>=0) close(Drives[0]);
  if(Drives[1]>=0) close(Drives[1]);
}



/****************************************************************/
/*** Load a cartridge ROM from .ROM file into given slot.     ***/
/****************************************************************/
int LoadCart(char *Name,int Slot)
{
  static char *ROMNames[] = 
  { "KONAMI/8kB","KONAMI/16kB","ASCII/8kB","ASCII/16kB" };
  static char Text[] = "  Cartridge # ROM #";
  FILE *F;
  int C1,C2,C3,L;

  if(!(F=fopen(Name,"rb"))) return(0);
  if(Verbose) printf("Found %s:\n",Name);

  C1=fgetc(F);C2=fgetc(F);
  if((C1!='A')||(C2!='B'))
  { 
    if(Verbose) printf("  Not a valid cartridge ROM\n");
    fclose(F);return(0);
  }

  C2=fgetc(F);C3=C2=(fgetc(F)>>6)&0x03;
  for(rewind(F);C2<3;C2++)
  {
    C1=fgetc(F);
    if(C1<0) break; else ungetc(C1,F);
    Text[12]='0'+Slot;Text[18]='0'+C2;
    LoadROM(Text,F,Slot,0,C2);
  }
  C1=fgetc(F);

  if((C3==2)&&(C2==3)&&(C1>=0))
  {
    ungetc(C1,F);
    MemMap[Slot][0][1]=MemMap[Slot][1][1]=
    MemMap[Slot][2][1]=MemMap[Slot][3][1]=MemMap[Slot][0][2];
    MemMap[Slot][0][2]=MemMap[Slot][1][2]=
    MemMap[Slot][2][2]=MemMap[Slot][3][2]=NULL;
    LoadROM(Text,F,Slot,0,2);
    C1=fgetc(F);
  }

  if((C2==3)&&(C1>=0))
  {
    if(Verbose) printf("  Loading %s MegaROM blocks...",ROMNames[ROMType]);
    ungetc(C1,F);C1=Slot-1;
    for(C2=0;C2<4;C2++)
      if(!(ROMMap[C1][C2]=malloc(PAGESIZE/2))) break;
    if(C2<4)
    { fclose(F);PRINTFAILED;return(0); }

    memcpy(ROMMap[C1][0],MemMap[Slot][0][1],PAGESIZE/2);
    memcpy(ROMMap[C1][2],MemMap[Slot][0][2],PAGESIZE/2);
    memcpy(ROMMap[C1][1],MemMap[Slot][0][1]+PAGESIZE/2,PAGESIZE/2);
    memcpy(ROMMap[C1][3],MemMap[Slot][0][2]+PAGESIZE/2,PAGESIZE/2);

    for(Slot--;C2<256;C2++)
    {
      C1=fgetc(F);
      if(C1<0) break; else ungetc(C1,F);
      if(!(ROMMap[Slot][C2]=malloc(PAGESIZE/2))) break;
      if(fread(ROMMap[Slot][C2],1,PAGESIZE/2,F)!=PAGESIZE/2)
      { free(ROMMap[Slot][C2]);ROMMap[Slot][C2]=NULL;break; }
    }

    if(Verbose) printf("%d blocks loaded\n",C2/(ROMType&1? 2:1));
  }
  
  fclose(F);return(1);
}



/****************************************************************/
/*** Load a ROM page from the file F into slot P, subslot S,  ***/
/*** page A. Return memory address of a page or 0 if failed.  ***/
/****************************************************************/
byte *LoadROM(char *STR,FILE *F,int P,int S,int A)
{
  if(Verbose) printf("%s: allocating memory...",STR);
  if(MemMap[P][S][A]) free(MemMap[P][S][A]);
  if(MemMap[P][S][A]=malloc(PAGESIZE))
  {
    if(Verbose) printf("loading...");
    if(fread(MemMap[P][S][A],1,PAGESIZE,F)!=PAGESIZE)
    { free(MemMap[P][S][A]);MemMap[P][S][A]=NULL; }
  }
  if(P!=3) 
    MemMap[P][0][A]=MemMap[P][1][A]=
    MemMap[P][2][A]=MemMap[P][3][A]=MemMap[P][S][A];
  if(Verbose) printf(MemMap[P][S][A]? MsgOK:MsgFAILED);  
  return(MemMap[P][S][A]);
}



/****************************************************************/
/*** Write number into given VDP register.                    ***/
/****************************************************************/
void VDPOut(register byte R,register byte V)
{ 
  register byte I,J,*P;  

  switch(R)  
  {
    case  0: switch(((V&0x0E)>>1)|(VDP[1]&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
               default:   J=ScrMode;
             }   
             if(J!=ScrMode)
             {
               I=((J>6)&&(J!=MAXSCREEN+1))? 11:10;
               ChrTab=VRAM+((long)(VDP[2]&SCR[J].R2)<<I);
               ChrGen=VRAM+((long)(VDP[4]&SCR[J].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&SCR[J].R3)<<6)
                          +((long)VDP[10]<<14);
	       SprTab=VRAM+((long)(VDP[5]&SCR[J].R5)<<7)
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               SCR[ScrMode=J].Init();SCR[J].Colors();
             }
             break;
    case  1: switch(((VDP[0]&0x0E)>>1)|(V&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
               default:   J=ScrMode;
             }   
             if(J!=ScrMode)
             {
               I=((J>6)&&(J!=MAXSCREEN+1))? 11:10;
               ChrTab=VRAM+((long)(VDP[2]&SCR[J].R2)<<I);
               ChrGen=VRAM+((long)(VDP[4]&SCR[J].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&SCR[J].R3)<<6)
                          +((long)VDP[10]<<14);
               SprTab=VRAM+((long)(VDP[5]&SCR[J].R5)<<7)  
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               SCR[ScrMode=J].Init();SCR[J].Colors();
             }
             break;       
    case  2: J=((ScrMode>6)&&(ScrMode!=MAXSCREEN+1))? 11:10;
             ChrTab=VRAM+((long)(V&SCR[ScrMode].R2)<<J);
             break;
    case  3: ColTab=
               VRAM+((long)(V&SCR[ScrMode].R3)<<6)+((long)VDP[10]<<14);
             break;
    case  4: ChrGen=VRAM+((long)(V&SCR[ScrMode].R4)<<11);
             break;
    case  5: SprTab=
               VRAM+((long)(V&SCR[ScrMode].R5)<<7)+((long)VDP[11]<<15);
             break;
    case  6: V&=0x3F;SprGen=VRAM+((long)V<<11);
             break;
    case  7: FGColor=V>>4;BGColor=V&0x0F;
             SCR[ScrMode].Colors();
             break;
    case  8:
    case  9: VDP[R]=V;SCR[ScrMode].Colors();break;
    case 10: V&=0x07;
             ColTab=
               VRAM+((long)(VDP[3]&SCR[ScrMode].R3)<<6)+((long)V<<14);
             break;
    case 11: V&=0x03;
             SprTab=
               VRAM+((long)(VDP[5]&SCR[ScrMode].R5)<<7)+((long)V<<15);
             break;
    case 14: V&=0x07;VPAGE=(V<VRAMPages)? VRAM+((long)V<<14):NULL;
             break;
    case 15: V&=0x0F;break;
    case 16: V&=0x0F;break;
    case 17: V&=0xBF;break;
    case 33:
    case 37:
    case 41: V&=0x01;break;
    case 35:
    case 39:
    case 43: V&=0x03;break;
    case 44: if(VDPStatus[2]&0x80)
             {
               if(MMC.Mode)
               {
                 if(V||MMC.TR)
                 {
                   P=MMC.P+MMC.XC/MMC.Diff;
                   J=MMC.Shift[MMC.XC%MMC.Diff];
                   *P=(*P&~(MMC.Mask<<J))|Oper(V<<J,*P,MMC.Op);
                 }
                 MMC.NXC--;
               }
               else
               {
                 P=MMC.P+MMC.XC/MMC.Diff;
                 *P=V;
                 MMC.NXC-=MMC.Diff;
               }
               if(MMC.NXC>0) MMC.XC+=MMC.XStep;
               else
                 if(--MMC.NY)
                 { MMC.XC=MMC.X;MMC.NXC=MMC.NX;MMC.P+=MMC.YStep; }
                 else VDPStatus[2]&=0x7E;
             }
             break;
    case 46: if((ScrMode>4)&&(ScrMode<9)) DoVDP(V);
             break;
  }
  VDP[R]=V;return;
} 



/****************************************************************/
/*** This function is used by DoVDP() for operations on       ***/
/*** pixels.                                                  ***/
/****************************************************************/
byte Oper(register byte A,register byte B,register byte N)
{
  switch(N)
  {
    case 0:  return(A);
    case 1:  return(A&B);
    case 2:  return(A|B);
    case 3:  return(A^B);
    case 4:  return(~A);
  }
  return(B);
}



/****************************************************************/
/*** Perform given V9938 operation.                           ***/
/****************************************************************/
void DoVDP(register byte V)
{
  int SX,SY,DX,DY,NX,NY;
  register int TX,TY,L;
  register byte *P,SM,I,J,TR;

  static byte Shift[4][4] = { {4,0,0,0},{6,4,2,0},{4,0,0,0},{0,0,0,0} };
  static byte Diff[4]     = { 2,4,2,1 };
  static byte Mask[4]     = { 0x0F,0x03,0x0F,0xFF };
  static int  Width[4]    = { 256,512,512,256 };
  static byte Line[4]     = { 7,7,8,8 };
  static int  Bytes[4]    = { 128,128,256,256 };

  SX=VDP[32]+((word)VDP[33]<<8);SY=VDP[34]+((word)VDP[35]<<8);
  DX=VDP[36]+((word)VDP[37]<<8);DY=VDP[38]+((word)VDP[39]<<8);
  NX=VDP[40]+((word)VDP[41]<<8);NY=VDP[42]+((word)VDP[43]<<8);

  SM=ScrMode-5;TR=!(V&0x08);

  if(Verbose&0x02)
    printf
    (
      "VDP COMMAND: %X (%d,%d)->(%d,%d) [%d,%d] Color %d\n",
      V,SX,SY,DX,DY,VDP[45]&0x04? -NX:NX,VDP[45]&0x08? -NY:NY,VDP[44]
    );

  switch(V&0xF0)
  {
    /*** Abort current operation: ***/
    case 0x00: VDPStatus[2]&=0x7E;
               break;

    /*** Get color of a point at SX,SY: ***/
    case 0x40: P=VRAM+((long)SY<<Line[SM])+SX/Diff[SM];
               VDPStatus[7]=(*P>>Shift[SM][SX%Diff[SM]])&Mask[SM];
               break;

    /*** Set color of a point at DX,DY: ***/
    case 0x50: J=VDP[44]&Mask[SM];
               if(J||TR)
               {
                 V&=0x07;
                 P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                 I=Shift[SM][DX%Diff[SM]];
                 *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
               }
               break;

    /*** Search in given dir. for a point of given color from SX,SY: ***/
    case 0x60: {
                 register byte *T,K;

                 K=VDP[44]&Mask[SM];
                 P=VRAM+((long)SY<<Line[SM]);
                 I=(VDP[45]&0x02)>>1;

                 if(VDP[45]&0x04)
                   for(;SX>=0;SX--)
                   {                   
                     T=P+SX/Diff[SM];
                     J=(*T>>Shift[SM][SX%Diff[SM]])&Mask[SM];
                     if((J==K)||((J!=BGColor)^I)) break;
                   }
                 else
                   for(;SX<Width[SM];SX++)
                   {
                     T=P+SX/Diff[SM];
                     J=(*T>>Shift[SM][SX%Diff[SM]])&Mask[SM];
                     if((J==K)||((J!=BGColor)^I)) break;
                   }
                 SX=(SX>=Width[SM])? Width[SM]-1:(SX<0)? 0:SX;
                 VDPStatus[8]=SX&0xFF;VDPStatus[9]=(SX>>8)|0xFE;
                 if(J==VDP[44]) VDPStatus[2]&=0xEF;
                 else           VDPStatus[2]|=0x10;
               }
               break;

    /*** Draw a line of length NX,NY from DX,DY: ***/
    case 0x70: J=VDP[44]&Mask[SM];
               if(J||TR)
               {
                 V&=0x07;
                 if(!NX&&!NY)
                 {
                    P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                    I=Shift[SM][DX%Diff[SM]];
                    *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V); 
                 }
                 else
                 {
                   register int AX,AY;

                   if(VDP[45]&0x01) { L=NX;NX=NY;NY=L; }
                   TX=VDP[45]&0x04? -1:1;TX=NX? TX:0;SX=DX+(NX+1)*TX;AX=2*NX;
                   TY=VDP[45]&0x08? -1:1;TY=NY? TY:0;SY=DY+(NY+1)*TY;AY=2*NY;

                   if(VDP[45]&0x01)
                     for(L=AX-NY;DY!=SY;DY+=TY,L+=AX)
                     {
                       P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                       I=Shift[SM][DX%Diff[SM]];
                       *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                       if(L>=0) { L-=AY;DX+=TX; }
                     }
                   else
                     for(L=AY-NX;DX!=SX;DX+=TX,L+=AY)
                     {
                       P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                       I=Shift[SM][DX%Diff[SM]];
                       *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                       if(L>=0) { L-=AX;DY+=TY; }
                     }
                 }
               }
               break;                

    /*** Fill a box with a given color: ***/
    case 0xC0: {
                 register int X,M;
                 register byte *DP;

                 J=VDP[44]&Mask[SM];
                 J|=(J<<Shift[SM][0])|(J<<Shift[SM][1])|(J<<Shift[SM][2]);
                 if(!NX) NX=Width[SM];

                 SX=VDP[45]&0x04? DX-NX+1:DX;
                 SY=VDP[45]&0x08? DY-NY+1:DY;
                 DP=VRAM+((long)SY<<Line[SM]);

                 for(;NY;NY--,DP+=Bytes[SM])
                   for(X=SX,L=NX;L;)
                   {
                     I=X%Diff[SM];P=DP+X/Diff[SM];
                     if(I||(L<Diff[SM]))
                     { 
                       I=Mask[SM]<<Shift[SM][I];
                       *P=(*P&~I)|(J&I);L--;X++;
                     }
                     else
                     {
                       memset(P,J,L/Diff[SM]);
                       M=L%Diff[SM];X+=L-M;L=M;
                     } 
                   }
               }
               break;

    case 0x80: J=VDP[44]&Mask[SM];
               if(J||TR)
               {
                 register byte *DP;
                 
                 V&=0x07;
                 if(!NX) NX=Width[SM];
                 SX=VDP[45]&0x04? DX-NX+1:DX;
                 SY=VDP[45]&0x08? DY-NY+1:DY;
                 DP=VRAM+((long)SY<<Line[SM]);

                 for(DX=SX+NX;NY;NY--,DP+=Bytes[SM])
                   for(L=SX;L<DX;L++)
                   {
                     P=DP+L/Diff[SM];I=Shift[SM][L%Diff[SM]];
                     *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                   }
               }         
               break;

    /*** Copy a box from one place to another: ***/
    case 0xE0: {
                 register int X,M;
                 register byte *SP,*DP;

                 NX=VDP[45]&0x08? DX:Width[SM]-DX;
                 TX=VDP[45]&0x04? -1:1;
                 TY=VDP[45]&0x08? -Bytes[SM]:Bytes[SM];
                 SP=VRAM+((long)SY<<Line[SM]);
                 DP=VRAM+((long)DY<<Line[SM]);   

                 for(;NY;NY--,SP+=TY,DP+=TY)
                   for(L=NX,X=SX;L;)
                   {
                     J=X%Diff[SM];P=SP+X/Diff[SM];
                     if(J||(L<Diff[SM]))
                     {
                       I=Shift[SM][J];
                       J=(*P>>I)&Mask[SM];
                       P=DP+X/Diff[SM];
                       *P=(*P&~(Mask[SM]<<I))|(J<<I);
                       L--;X+=TX;
                     }
                     else
                     {
                       M=L/Diff[SM];
                       if(TX>0)
                         memcpy(DP+X/Diff[SM],P,M);
                       else
                         memcpy(DP+X/Diff[SM]-M+1,P-M+1,M);
                       M=L%Diff[SM];X+=TX*(L-M);L=M;
                     }
                   }
               }   
               break;

    case 0xD0: {
                 register int XS,XD,M;
                 register byte *SP,*DP;

                 if(!NX) NX=Width[SM];
                 TX=VDP[45]&0x04? -1:1;
                 TY=VDP[45]&0x08? -Bytes[SM]:Bytes[SM];
                 SP=VRAM+((long)SY<<Line[SM]);
                 DP=VRAM+((long)DY<<Line[SM]);

                 for(;NY;NY--,SP+=TY,DP+=TY)
                   for(L=NX,XS=SX,XD=DX;L;)
                   {
                     J=XS%Diff[SM];I=XD%Diff[SM];P=SP+XS/Diff[SM];
                     if(J||I||(L<Diff[SM]))
                     {
                       J=(*P>>Shift[SM][J])&Mask[SM];
                       P=DP+XD/Diff[SM];I=Shift[SM][I];
                       *P=(*P&~(Mask[SM]<<I))|(J<<I);
                       L--;XS+=TX;XD+=TX;
                     }
                     else
                     {
                       M=L/Diff[SM];
                       if(TX>0)
                         memcpy(DP+XD/Diff[SM],P,M);
                       else
                         memcpy(DP+XD/Diff[SM]-M+1,P-M+1,M);
                       M=L%Diff[SM];L=TX*(L-M);XS+=L;XD+=L;L=M;
                     }
                   }
               }   
               break;

    case 0x90: {
                 register int XS,XD;
                 register byte *SP,*DP;
                       
                 V&=0x07;
                 if(!NX) NX=Width[SM];
                 TX=VDP[45]&0x04? -1:1;
                 TY=VDP[45]&0x08? -Bytes[SM]:Bytes[SM];
                 SP=VRAM+((long)SY<<Line[SM]);
                 DP=VRAM+((long)DY<<Line[SM]);

                 for(;NY;NY--,SP+=TY,DP+=TY)
                   for(L=NX,XS=SX,XD=DX;L;L--,XS+=TX,XD+=TX)
                   {
                     P=SP+XS/Diff[SM];
                     J=(*P>>Shift[SM][XS%Diff[SM]])&Mask[SM];
                     if(J||TR)
                     {                   
                       P=DP+XD/Diff[SM];I=Shift[SM][XD%Diff[SM]];
                       *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                     }
                   }
               }
               break;

    /*** Transfer a box of data from CPU: ***/
    case 0xF0: V&=0xF0;
    case 0xB0: VDPStatus[2]|=0x81;
               if(!NX) NX=Width[SM];
               MMC.Mode=((V&0xF0)==0xB0);
               MMC.TR=!(V&0x08);MMC.Op=V&0x07;
               MMC.P=VRAM+((long)DY<<Line[SM]);
               MMC.YStep=(VDP[45]&0x08? -1:1)<<Line[SM];
               MMC.XStep=(MMC.Mode? 1:Diff[SM])*(VDP[45]&0x04? -1:1);
               MMC.X=MMC.XC=DX;MMC.NX=MMC.NXC=NX;MMC.NY=NY;
               MMC.Diff=Diff[SM];MMC.Mask=Mask[SM];
               memcpy(MMC.Shift,Shift[SM],4);
               VDPOut(44,VDP[44]);
               break;

    default: printf("Unimplemented V9938 operation %X\n",V);
  }
}



/****************************************************************/
/*** Send a character to the printer.                         ***/
/****************************************************************/
void Printer(byte V) { fputc(V,PrnStream); }



/****************************************************************/
/*** Read number from given RTC register.                     ***/
/****************************************************************/
byte RTCIn(register byte R)
{
  register byte J;
  time_t T;
  char *S;

  /*** ctime() returns "DDD MMM dd hh:mm:ss YYYYn0" ***/
  static byte RTable[13] =
  { 18,17,15,14,12,11,25,9,8,25,25,23,22 };
  static char *Days[7] =
  { "Sun","Mon","Tue","Wed","Thu","Fri","Sat" }; 
  static char *Months[12] =
  {
    "Jan","Feb","Mar","Apr","May","Jun",
    "Jul","Aug","Sep","Oct","Nov","Dec"
  };

  J=RTCMode&0x03;
  if(R>12) J=(R==13)? RTCMode:NORAM;
  else
    if(J) J=RTC[J][R];
    else
    {
      time(&T);S=ctime(&T);
      switch(R)
      {
        case 6:  for(J=0;(J<7)&&strncmp(S,Days[J],3);J++);
                 J=(J==7)? 0:J+1;break;
        case 9:  for(J=0;(J<12)&&strncmp(S+4,Months[J],3);J++);
                 J=(J==12)? 0:(J+1)%10;break;
        case 10: for(J=0;(J<12)&&strncmp(S+4,Months[J],3);J++);
                 J=(J==12)? 0:(J+1)/10;break;
        case 11: S[24]='\0';J=(atoi(S+20)-1980)%10;break;
        case 12: S[24]='\0';J=((atoi(S+20)-1980)/10)%10;break;
        default: J=S[RTable[R]]-'0';
      }
    }
  return(J|0xF0);
}



/****************************************************************/
/*** Set sound parameters, refresh screen, check keyboard and ***/
/*** sprites. Call this function on each interrupt.           ***/
/****************************************************************/
void Interrupt(void)
{
  static byte BFlag=0;
  static byte BCount=0;
  static int UCount=1;
  register byte I,J;

  /*** Checking keyboard: ***/
  Keyboard();

  if((!--UCount)||!EndOfFrame)
  {
    UCount=UPeriod;
  
    /*** Blinking for TEXT80: ***/
    if(BCount) BCount--;
    else
    {
      BFlag=!BFlag;
      if(VDP[13])
      {
        BCount=BFlag? VDP[13]&0x0F:VDP[13]>>4;
        if(BCount)
          if(BFlag)
          { XFGColor=FGColor;XBGColor=BGColor; }
          else 
          { XFGColor=VDP[12]>>4;XBGColor=VDP[12]&0x0F; }
      }  
      else { XFGColor=FGColor;XBGColor=BGColor; }
    }

    /*** Refreshing screen: ***/
    J=VDP[19]-VDP[23];
    I=ScanLines212? 211:191;
    if((ScanLine>J)||(J>I)||!(VDP[0]&0x10)) J=I;
    else VDPStatus[1]|=0x01;
    EndOfFrame=(J==I);
    (SCR[ScrMode].Refresh)(ScanLine,J);
    ScanLine=(J<I)? J+1:0;

    /*** Checking sprites: ***/
    if(!SpritesOFF&&ScrMode&&(ScrMode<MAXSCREEN+1)) CheckSprites();
  }

  /*** Generating VDP interrupt: ***/
  if((VDP[1]&0x20)&&EndOfFrame) VDPStatus[0]|=0x80;
}



/****************************************************************/
/*** Read/Write sectors to a disk. This function is called on ***/
/*** ED FE instruction to emulate disk access.                ***/
/****************************************************************/
void Disk(reg *R)
{
  static struct
  { int Sectors;byte Heads,Names,PerTrack,PerFAT,PerCluster; }
  Info[8] =
  {
    {  720,1,112,9,2,2 },
    { 1440,2,112,9,3,2 },
    {  640,1,112,8,1,2 },
    { 1280,2,112,8,2,2 },
    {  360,1, 64,9,2,1 },
    {  720,2,112,9,2,2 },
    {  320,1, 64,8,1,1 },
    {  640,2,112,8,1,2 }
  };

  byte *A,Buf[512],J,PS,SS;
  int F,N,M;

  switch(R->PC.W-2)
  {
    /*** Read/write sectors from disk: ***/
    case 0x4010:
      if(Verbose&0x04)
        printf
        (
          "DISK %s: %d sectors starting from %Xh [buffer at %Xh]\n",
          R->AF.B.l&C_FLAG? "WRITE":"READ",R->BC.B.h,R->DE.W,R->HL.W
        );

      A=RAM+R->HL.W;J=R->BC.B.h;R->BC.B.h=0;

      if((F=Drives[R->AF.B.h&0x01])<0)
      { R->AF.W=0x0201;return; }
      lseek(F,0x15,0);read(F,Buf,1);
      if(R->BC.B.l!=*Buf)
      { R->AF.W=0x1201;return; }
      if(
          (R->DE.W>=Info[R->BC.B.l-0xF8].Sectors)||
          (lseek(F,R->DE.W*512,0)!=R->DE.W*512)
        )
      { R->AF.W=0x0601;return; }

      PS=PSLReg;SS=SSLReg;
      DoOut(0xA8,0xFF);SSlot(0xAA);

      if(R->AF.B.l&C_FLAG)
        while(J--)
          if(write(F,A,512)!=512)
          { R->AF.W=0x1001;SSlot(SS);DoOut(0xA8,PS);return; }
          else
          { A+=512;R->BC.B.h++; }
      else
        while(J--)
          if(read(F,A,512)!=512)
          { R->AF.W=0x0801;SSlot(SS);DoOut(0xA8,PS);return; }
          else
          { A+=512;R->BC.B.h++; }

      SSlot(SS);DoOut(0xA8,PS);
      R->AF.W=0x0000;return;

    case 0x4013:
      if((F=Drives[R->AF.B.h&0x01])<0)
      { R->AF.W=0x0201;return; }

      lseek(F,0x15,0);read(F,Buf,1);
      R->BC.B.h=(R->BC.B.l==*Buf)? 0x01:0xFF; 
      R->AF.B.l&=~C_FLAG;return;

    /*** Format disk: ***/
    case 0x401C:
      J=R->DE.B.h&0x01;
      if(Drives[J]>=0) close(Drives[J]);
      F=Drives[J]=open(J? DiskB:DiskA,O_RDWR|O_CREAT,0777);
      if(!F) { R->AF.W=0x0201;return; }      

      lseek(F,0,0);J=(2-R->AF.B.h)&0x01;

      A=BootBlock+3;
      memcpy(A,"fMSXdisk",8);A+=10;                /* Manufacturer's ID   */
      *A=Info[J].PerCluster;A+=4;                  /* Sectors per cluster */
      *A++=Info[J].Names;*A++=0x00;                /* Number of names     */
      *A++=Info[J].Sectors;*A++=0x00;              /* Number of sectors   */
      *A++=J+0xF8;                                 /* Format ID [F8h-FFh] */
      *A++=Info[J].PerFAT;*A++=0x00;               /* Sectors per FAT     */
      *A++=Info[J].PerTrack;*A++=0x00;             /* Sectors per track   */
      *A++=Info[J].Heads;*A=0x00;                  /* Number of heads     */
      if(write(F,BootBlock,512)!=512) { R->AF.W=0x1001;return; };

      Buf[0]=J+0xF8;Buf[1]=Buf[2]=0xFF;
      memset(Buf+3,0x00,509);
      for(N=0;N<2;N++)
      {
        Buf[0]=J+0xF8;Buf[1]=Buf[2]=0xFF;memset(Buf+3,0x00,509);
        if(write(F,Buf,512)!=512) { R->AF.W=0x1001;return; }
        memset(Buf,0x00,512);
        for(M=Info[J].PerFAT;M>1;M--)
          if(write(F,Buf,512)!=512)
          { R->AF.W=0x1001;return; }
      }

      N=Info[J].Names/16;memset(Buf,0x00,512);
      M=Info[J].Sectors-2*Info[J].PerFAT-N-1;
      for(;N;N--)
        if(write(F,Buf,512)!=512)
        { R->AF.W=0x1001;return; }      

      memset(Buf,0xFF,512);
      for(;M;M--)
        if(write(F,Buf,512)!=512)
        { R->AF.W=0x1001;return; }

      R->AF.B.l&=~C_FLAG;return;

    default:
      printf("Disk routine called at PC=%X\n",R->PC.W-2);
  }
}



/****************************************************************/
/*** Check for sprite collisions and 5th/9th sprite in a row. ***/
/****************************************************************/
void CheckSprites(void)
{
  register char DH,DV,SH,SV;
  register byte I,J,N,M,*PS,*PD,*S,*D;

  VDPStatus[0]&=0xB0;
  M=(ScrMode>3)? 8:4;

  if(Sprites16x16)
    for(J=0,S=SprTab;J<32;J++,S+=4)
    {
      N=0;
      for(I=J+1,D=S+4;I<32;I++,D+=4)
      {
        DV=S[0]-D[0];DV*=(SV=(DV<0)? -1:1);
        if(DV<16)
	{
          N++;DH=S[1]-D[1];DH*=(SH=(DH<0)? -1:1);
          if(DH<16)
	  {
            PS=SprGen+((long)(S[2]&0xFC)<<3);
            PD=SprGen+((long)(D[2]&0xFC)<<3);
            if(SV>0) PD+=DV; else PS+=DV;
            if(SH>0)
              while
              (
                (DV<16)&&!((long)*PS&((long)*PD<<DH))&&
                !((long)*(PS+16)&((long)*(PD+16)<<DH))
              ) { DV++;PS++;PD++; }
            else
              while
              (
                (DV<16)&&!((long)*PD&((long)*PS<<DH))&&
                !((long)*(PD+16)&((long)*(PS+16)<<DH))
              ) { DV++;PS++;PD++; }
            if(DV<16) VDPStatus[0]|=0x20;     
          }
        }
      }  
      if(M>4) VDPStatus[0]=(VDPStatus[0]&0xF0)|0x40|(J&0x0F);
    }
  else
    for(J=0,S=SprTab;J<32;J++,S+=4)
    {
      N=0;
      for(I=J+1,D=S+4;I<32;I++,D+=4)
      {
        DV=S[0]-D[0];DV*=(SV=(DV<0)? -1:1);
        if(DV<8)
        {
          N++;DH=S[1]-D[1];DH*=(SH=(DH<0)? -1:1);
          if(DH<8)
          {
            PS=SprGen+((long)S[2]<<3);PD=SprGen+((long)D[2]<<3);
            if(SV>0) PD+=DV; else PS+=DV;
            if(SH>0)
              while((DV<8)&&!((long)*PS&((long)*PD<<DH)))
              { DV++;PS++;PD++; }
            else
              while((DV<8)&&!((long)*PD&((long)*PS<<DH)))
              { DV++;PS++;PD++; }
            if(DV<8) VDPStatus[0]|=0x20;
          }
        }
      }
      if(M>4) VDPStatus[0]=(VDPStatus[0]&0xF0)|0x40|(J&0x0F);
    }
}
